package middleware

import (
	"context"
	"strings"

	"git.perx.ru/perxis/perxis-go/pkg/cache"
	"git.perx.ru/perxis/perxis-go/pkg/data"
	service "git.perx.ru/perxis/perxis-go/pkg/environments"
)

func makeKey(ss ...string) string {
	return strings.Join(ss, "-")
}

func CachingMiddleware(cache cache.Cache) Middleware {
	return func(next service.Environments) service.Environments {
		return &cachingMiddleware{
			cache: cache,
			next:  next,
		}
	}
}

type cachingMiddleware struct {
	cache cache.Cache
	next  service.Environments
}

func (m cachingMiddleware) Create(ctx context.Context, env *service.Environment) (created *service.Environment, err error) {

	created, err = m.next.Create(ctx, env)
	if err == nil {
		_ = m.cache.Remove(created.SpaceID)
	}
	return created, err
}

func (m cachingMiddleware) Get(ctx context.Context, spaceId string, id string) (env *service.Environment, err error) {

	value, e := m.cache.Get(makeKey(spaceId, id))
	if e == nil {
		return value.(*service.Environment).Clone(), nil
	}
	env, err = m.next.Get(ctx, spaceId, id)
	if err == nil {
		_ = m.cache.Set(makeKey(spaceId, env.ID), env)
		for _, a := range env.Aliases {
			_ = m.cache.Set(makeKey(spaceId, a), env)
		}
		return env.Clone(), nil
	}
	return nil, err
}

func (m cachingMiddleware) List(ctx context.Context, spaceId string) (envs []*service.Environment, err error) {

	value, e := m.cache.Get(spaceId)
	if e == nil {
		return data.CloneSlice(value.([]*service.Environment)), nil
	}
	envs, err = m.next.List(ctx, spaceId)
	if err == nil {
		_ = m.cache.Set(spaceId, envs)
		return data.CloneSlice(envs), nil
	}
	return nil, err
}

func (m cachingMiddleware) Update(ctx context.Context, env *service.Environment) (err error) {

	err = m.next.Update(ctx, env)
	if err == nil {
		value, e := m.cache.Get(makeKey(env.SpaceID, env.ID))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}
		_ = m.cache.Remove(env.SpaceID)
	}
	return err
}

func (m cachingMiddleware) Delete(ctx context.Context, spaceId string, id string) (err error) {

	err = m.next.Delete(ctx, spaceId, id)
	if err == nil {
		value, e := m.cache.Get(makeKey(spaceId, id))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}
		_ = m.cache.Remove(spaceId)
	}
	return err
}

func (m cachingMiddleware) SetAlias(ctx context.Context, spaceId string, id string, alias string) (err error) {

	err = m.next.SetAlias(ctx, spaceId, id, alias)
	if err == nil {
		value, e := m.cache.Get(makeKey(spaceId, alias))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}

		value, e = m.cache.Get(makeKey(spaceId, id))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}
		_ = m.cache.Remove(spaceId)
	}
	return err
}

func (m cachingMiddleware) RemoveAlias(ctx context.Context, spaceId string, id string, alias string) (err error) {

	err = m.next.RemoveAlias(ctx, spaceId, id, alias)
	if err == nil {
		_ = m.cache.Remove(spaceId)
		value, e := m.cache.Get(makeKey(spaceId, alias))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}

		value, e = m.cache.Get(makeKey(spaceId, id))
		if e == nil {
			env := value.(*service.Environment)
			_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
			for _, a := range env.Aliases {
				_ = m.cache.Remove(makeKey(env.SpaceID, a))
			}
		}
	}
	return err
}

func (m cachingMiddleware) Migrate(ctx context.Context, spaceId, id string, options ...*service.MigrateOptions) (err error) {
	err = m.next.Migrate(ctx, spaceId, id, options...)

	// значение из кэша удалить вне зависимости от наличия ошибки, поскольку состояние окружения могло измениться
	value, e := m.cache.Get(makeKey(spaceId, id))
	if e == nil {
		env := value.(*service.Environment)
		_ = m.cache.Remove(makeKey(env.SpaceID, env.ID))
		for _, a := range env.Aliases {
			_ = m.cache.Remove(makeKey(env.SpaceID, a))
		}
	}
	_ = m.cache.Remove(spaceId)
	return err
}
