package extension

import (
	"context"
	"strings"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/operation"
	commonpb "git.perx.ru/perxis/perxis-go/proto/common"
	pb "git.perx.ru/perxis/perxis-go/proto/extensions"
	"google.golang.org/grpc"
	"google.golang.org/protobuf/proto"
)

type RouteFn func(ctx context.Context, extensions ...string) ([]Extension, error)

type WrapErrFn func(extension Extension, err error) error

func DefaultWrapErrFn() WrapErrFn {
	return func(_ Extension, err error) error { return err }
}

type ServerOption func(c *Server)

// WrapErrs Оборачивать ошибки, возвращаемые от каждого из расширений, в название расширения
func WrapErrs() ServerOption {
	return func(c *Server) {
		c.wrapErr = func(extension Extension, err error) error {
			return errors.Wrap(err, extension.GetDescriptor().Extension)
		}
	}
}

// NamedExtensions Перенаправлять запросы к сервисам с названиями, соответствующими
// списку расширений в XXXRequest. Если не найден сервис с названием, возвращается ошибка
func NamedExtensions(extensions ...Extension) ServerOption {
	return func(c *Server) {
		m := make(map[string]Extension, len(extensions))
		for _, e := range extensions {
			m[e.GetDescriptor().Extension] = e
		}
		c.extensions = func(_ context.Context, extensions ...string) ([]Extension, error) {
			var res []Extension
			if len(extensions) == 0 {
				for _, e := range m {
					res = append(res, e)
				}
				return res, nil
			}

			for _, ext := range extensions {
				e, ok := m[ext]
				if !ok {
					return nil, errors.Wrap(ErrUnknownExtension, ext)
				}
				res = append(res, e)
			}
			return res, nil
		}
		c.wrapErr = func(extension Extension, err error) error {
			return errors.Wrap(err, extension.GetDescriptor().Extension)
		}
	}
}

// SingleExtension Перенаправлять запросы к единому сервису
func SingleExtension(extension Extension) ServerOption {
	return func(c *Server) {
		c.extensions = func(_ context.Context, _ ...string) ([]Extension, error) {
			return []Extension{extension}, nil
		}
		c.wrapErr = DefaultWrapErrFn()
	}
}

func WithRouteFn(fn RouteFn) ServerOption {
	return func(c *Server) {
		c.extensions = fn
	}
}

type Server struct {
	extensions RouteFn
	wrapErr    WrapErrFn
	operations operation.Service
	pb.UnimplementedExtensionServiceServer
}

func NewServer(options ...ServerOption) *Server {
	srv := &Server{
		operations: operation.NewDefaultService(),
		wrapErr:    DefaultWrapErrFn(),
	}

	for _, o := range options {
		o(srv)
	}

	return srv
}

func RegisterServer(grpcSrv *grpc.Server, srv *Server) {
	commonpb.RegisterOperationServiceServer(grpcSrv, operation.NewServer(srv.operations))
	pb.RegisterExtensionServiceServer(grpcSrv, srv)
}

func (s *Server) Install(ctx context.Context, req *InstallRequest) (*operation.Proto, error) {
	exts, err := s.extensions(ctx, req.Extensions...)
	if err != nil {
		return nil, err
	}

	desc := "Install extensions: " + strings.Join(req.Extensions, ", ")
	op, err := s.operations.Create(ctx, desc,
		func(ctx context.Context) (proto.Message, error) {
			for _, ext := range exts {
				if err := ext.Install(ctx, req); err != nil {
					return nil, s.wrapErr(ext, err)
				}
			}
			return nil, nil
		})

	return op.Proto(), err
}

func (s *Server) Uninstall(ctx context.Context, req *UninstallRequest) (*operation.Proto, error) {
	exts, err := s.extensions(ctx, req.Extensions...)
	if err != nil {
		return nil, err
	}

	desc := "Uninstall extensions: " + strings.Join(req.Extensions, ", ")
	op, err := s.operations.Create(ctx, desc,
		func(ctx context.Context) (proto.Message, error) {
			for _, ext := range exts {
				if err := ext.Uninstall(ctx, req); err != nil {
					return nil, s.wrapErr(ext, err)
				}
			}
			return nil, nil
		})

	return op.Proto(), err
}

func (s *Server) Check(ctx context.Context, req *CheckRequest) (*operation.Proto, error) {
	exts, err := s.extensions(ctx, req.Extensions...)
	if err != nil {
		return nil, err
	}

	desc := "Check extensions: " + strings.Join(req.Extensions, ", ")
	op, err := s.operations.Create(ctx, desc,
		func(ctx context.Context) (proto.Message, error) {
			for _, ext := range exts {
				if err := ext.Check(ctx, req); err != nil {
					return nil, s.wrapErr(ext, err)
				}
			}
			return nil, nil
		})

	return op.Proto(), err
}

func (s *Server) Action(ctx context.Context, in *pb.ActionRequest) (*pb.ActionResponse, error) {
	actionURL, err := NewActionURL(in.Action)
	if err != nil {
		return nil, ErrInvalidAction
	}

	ext := actionURL.Extension()
	if ext == "" {
		ext = in.Extension
	}
	if ext == "" {
		return nil, ErrInvalidAction
	}

	svc, err := s.extensions(ctx, ext)
	if err != nil {
		return nil, err
	}

	out, err := svc[0].Action(ctx, ActionRequestFromPB(in))

	if out == nil {
		out = &ActionResponse{}
	}

	if err != nil {
		out.State = ResponseError
		out.Error = err.Error()
		out.Msg += errors.GetDetail(err)
	}

	return ActionResponseToPB(out), nil
}

func (s *Server) Start() error {
	extensions, _ := s.extensions(context.Background())
	var errs []error
	for _, svc := range extensions {
		if r, ok := svc.(Runnable); ok {
			if err := r.Start(); err != nil {
				errs = append(errs, err)
			}
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrStart, errs...)
	}

	return nil
}

func (s *Server) Stop() error {
	extensions, _ := s.extensions(context.Background())
	var errs []error
	for _, svc := range extensions {
		if r, ok := svc.(Runnable); ok {
			if err := r.Stop(); err != nil {
				errs = append(errs, err)
			}
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrStop, errs...)
	}

	return nil
}
